import libvirt
import libxml2

from vminstall.exceptions import *
import vminstall.xml
import vminstall.paths
import vminstall.util
import vminstall.caps
from . import disks

class DiskMapStatic:
    def __init__(self):
        self._modified = False
        self.disk_map = {}
        
        lines = []
        try:
            # FIXME: advisory locking
            f = open(vminstall.paths.imagedir + '.disk-list')
            lines = f.readlines() 
            f.close()
        except:
            pass
        for line in lines:
            try:
                disk = disks.parse_string(line, extended=True)
            except:
                self._modified = True
                continue
            self._modified |= (not self._add_disk(disk))

    def _del_disk(self, pdev):
        del self.disk_map[pdev]
        self._modified = True
        
    def _add_disk(self, disk, vm=None):
        """Returns True if added; False if not added because already exists
        """
        if not disk.exists():
            return False
        pdev = disk.get_pdev()
        if pdev not in self.disk_map:
            self.disk_map[pdev] = []
        for map in self.disk_map[pdev]:
            if ((not vm and not map[1]) or
                (vm and map[1] and vm.uuid == map[1].uuid)):
                return False
        if vm:
            self.disk_map[pdev] = [m for m in self.disk_map[pdev] if m[1] is not None]
        self.disk_map[pdev].append((disk, vm))
        return True

    def _get_unused_disk(self, pdev):
        if pdev not in self.disk_map:
            raise DiskResourceError(DiskResourceError.E_NOT_FOUND, details=pdev)
        uses = self.disk_map[pdev]
        if len(uses) != 1 or uses[0][1] is not None:
            raise DiskResourceError(DiskResourceError.E_PERM, details=pdev)
        return self.disk_map[pdev][0][0]
        
    def manage(self, pdev):
        self.manage_disk(disks.parse_string(pdev))
        
    def manage_disk(self, disk):
        pdev = disk.get_pdev()
        self._modified |= self._add_disk(disk)
        
    def unmanage(self, pdev):
        self._get_unused_disk(pdev)
        self._del_disk(pdev)
        
    def unmanage_disk(self, disk):
        self.unmanage(disk.get_pdev())
        
    def delete(self, pdev):
        disk = self._get_unused_disk(pdev)
        try:
            disk.delete()
        except NotImplementedError:
            pass
        self._del_disk(pdev)
        
    def flush(self):
        if not self._modified:
            return
        # FIXME: advisory locking
        f = open(vminstall.paths.imagedir + '.disk-list', 'w')
        for maps in list(self.disk_map.values()):
            for map in maps:
                f.write(map[0].__repr__())
                f.write('\n')
        f.close()

    def __del__(self):
        try:
            self.flush()
        except:
            # Exceptions can't be thrown from destructor
            pass

class DiskMap(DiskMapStatic):
    def __init__(self):
        DiskMapStatic.__init__(self)
        
        class VM:
            def __repr__(self):
                return "'%s'" % self.uuid
        
        if vminstall.caps.is_xen:
            c = libvirt.openReadOnly('xen:///')
        else:
            c = libvirt.openReadOnly('qemu:///system')
        vm_names = c.listDefinedDomains() + c.listDomainsID()
        for vm_name in vm_names:
            if vm_name == 0:
                continue
            if type(vm_name) is int:
                libvirt_vm = c.lookupByID(vm_name)
            else:
                libvirt_vm = c.lookupByName(str(vm_name))
            vm_xml = libvirt_vm.XMLDesc(0)
            doc = libxml2.parseDoc(vm_xml)
            ctx = doc.xpathNewContext()

            vm = VM()
            vm.name = libvirt_vm.name()
            vm.uuid = vminstall.util.uuidToString([ord(a) for a in tuple(libvirt_vm.UUID())])
            for disk in vminstall.xml.get_disks(ctx.xpathEval('/domain/devices')[0]):
                self._modified |= self._add_disk(disk, vm)
